clc; clear;


%% Model equations

N_total = 60;

N_switching = 25;  % change in reference value

nx = 2;
ny = 2;
nu = 1;
nd = 2;

A = [0.99, -0.1;
     0.01   0.5];
B = [1; 1;];

C = [1 0;
     0 1;];
 
r_ref_1 = 2;
y_ref_1 = [2; 1];
r_ref_2 = 4;
y_ref_2 = [4; 1];

H = [1 0];

% covariances for state estimator
sigma_w = 0.1*eye(nx);
sigma_w_inv = inv(sigma_w);

sigma_v = 0.2*eye(ny);

% weighting matrices

Q = 0.5*eye(nx); 
R = 1;

Q_s = eye(ny);
R_s = 0.1;

% horizon
N_mpc = 5;

%% Plant

plant = Plant(A, B, C);

%% Closed-loop simulation with disturbance model

% init estimates
x_est = zeros(nx, 1);
P_est = eye(nx);

% init plant
plant.init();

% store results
x_est_traj = zeros(nx, N_total);
x_true_traj = zeros(nx, N_total);

u_traj = zeros(nu, N_total);
y_traj = zeros(ny, N_total);



for n=1:N_total
   if n < N_switching
       r_ref = r_ref_1;
       y_ref = y_ref_1;
   else
       r_ref = r_ref_2;
       y_ref = y_ref_2;
   end
   x_est_traj(:, n) = x_est;
   
   % get steady state
   [x_s, u_s] = getSteadyState(A, B, C, Q_s, R_s, y_ref, H, r_ref);
   
   % get next control input 
   u = getControl(x_est, x_s, u_s, A, B, Q, R, N_mpc);
   
   % apply the control to the plant and obtain the next measurement
   [y, x_true] = plant.step(u);
   
   % save results
   u_traj(:, n) = u;
   y_traj(:, n) = y;
   x_true_traj(:, n) = x_true;
   
   % estimate current state
   [x_est, P_est] = estimate(x_est, P_est, u, y, A, B, C, sigma_w_inv, sigma_v);

end


%% Augmented model with disturbances

B_d = [1, 0;
       0, 1];
C_d = [0, 0;
       0, 0;];

A_tilde = [A B_d;
           0 0  1 0;
           0 0  0 1];
B_tilde = [B;
           0;
           0;];

C_tilde = [C C_d];

sigma_w_tilde = blkdiag(sigma_w, 0.5*eye(nd));
sigma_w_tilde_inv = inv(sigma_w_tilde);

% Check detectability

rank([eye(nx)-A -B_d;
        C        C_d;])


%% Closed-loop simulation with disturbance model

% init plant
plant.init();

% init estimator
x_est = zeros(nx, 1);
d_est = zeros(nd, 1);

x_tilde_est = [x_est; d_est];
P_tilde_est = eye(nx+nd);

x_est_traj_d = zeros(nx, N_total);
d_est_traj = zeros(nd, N_total);

x_true_traj_d = zeros(nx, N_total);

u_traj_d = zeros(nu, N_total);
y_traj_d = zeros(ny, N_total);


for n=1:N_total
   
   if n < N_switching
       r_ref = r_ref_1;
       y_ref = y_ref_1;
   else
       r_ref = r_ref_2;
       y_ref = y_ref_2;
   end
   
   x_est_traj_d(:, n) = x_est;
   d_est_traj(:, n) = d_est;
   
   % get steady state
   [x_s, u_s] = getSteadyStateWithDisturbanceModel(A, B, B_d, C, C_d, Q_s, R_s, y_ref, H, r_ref, d_est);
   
   % get next control input 
   u = getControl(x_est, x_s, u_s, A, B, Q, R, N_mpc);
   
   % apply the control to the plant and obtain the next measurement
   [y, x_true] = plant.step(u);
   
   u_traj_d(:, n) = u;
   y_traj_d(:, n) = y;
   x_true_traj_d(:, n) = x_true;
   
   % update the state and disturbance estimate
   [x_tilde_est, P_tilde_est] = estimate(x_tilde_est, P_tilde_est, u, y, A_tilde, B_tilde, C_tilde, sigma_w_tilde_inv, sigma_v);   %x_est = x_tilde_est(1:nx);
   
   x_est = x_tilde_est(1:nx);
   d_est = x_tilde_est(nx+1:end);
end

